from .tsn import tsn_to_tuple
from .backend import get_backend_by_name, get_backend_for_tensor
from .transforms import _view_transform, _permute_transform, _join_transform, expand_transform

def get_backend(backend, x):
    if backend is not None:
        be = get_backend_by_name(backend)
    else:
        be = get_backend_for_tensor(x)

    return be



def join (tlist, dims, backend=None):
    '''
    tlist: List[tensor], list of tensors
    dims: str = '..,^,...' or '..,*,...' 
    '''
    assert isinstance(tlist, list), "Can only group a list of tensors."
    assert len(tlist) > 1, "Can only group more than one tensors."

    be = get_backend(backend, tlist[0])

    if len(dims) > 1: assert ',' in dims, 'Separate dimensions by "," here.'

    out_shape = dims.strip().split(',')
    if '^' in out_shape:
        pos = out_shape.index('^')
        return be.stack(tlist, axis=pos)
    else:
        if '*' not in out_shape:
            assert pos != '-1', 'Group specification does not contain "^" or "*".'

        pos = out_shape.index('*')
        return be.concat(tlist, axis=pos)


def tfm_seq_decompose (tfms, tfm_names):
    '''
    Decompose a multi-step transform into basic (view, permute, expand) transforms
    tfms  'btd -> b,t,2,d//2 -> b,2,t,d//2'
    tfm_names 'vp' , i.e., view, then permute transform
    '''
    tfm_symbols = list(tfm_names) # ['v', 't']
    tfm_symbols_no_c = list(tfm_names.replace('c',''))
    tfm_list = [] # (trf symbol, trf_lhs, trf_rhs)

    if isinstance(tfms, str):

        shapes = [t.strip() for t in tfms.split('->')]
        assert len(shapes) >= 2, "Specify at least one transform, e.g., btd->dtb"
        assert len(tfm_symbols_no_c) == (len(shapes) - 1), "Num of transform descriptions and symbols do not match"

        shapes = [tsn_to_tuple(s) for s in shapes]
        #for i, (l, r) in enumerate(zip(shapes[:-1], shapes[1:])):
        #    tfm_list.append((tfm_symbols[i], l, r) )

        curr_shape_pos = 0 #count current tfm's position (handle implicit contiguous)
        for sym in tfm_symbols:
            #contiguous transform
            if sym == 'c': 
                tfm_list.append((sym, None, None))
            else:
                l, r = shapes[curr_shape_pos: curr_shape_pos+2]
                tfm_list.append((sym, l, r))
                curr_shape_pos += 1

    elif isinstance(tfms, list):
        assert len(tfms) == len(tfm_symbols_no_c), "Num transformations {0} != transform symbols {1}".format(len(tfms),len(tfm_symbols_no_c))
        assert len(tfms) > 0, "No transformations given. Specify at least one transformation"
        curr_pos = 0 #count current tfm's position (handle implicit contiguous)
        for sym in tfm_symbols:
            if sym == 'c':   #contiguous transform
                tfm_list.append((sym, None, None))
            else:
                l, r = [t.strip() for t in tfms[curr_pos].split('->')]
                tfm_list.append((sym, l, r))
                curr_pos += 1

    else:
        assert False, "warp: wrong format for transforms. Specify either a string or a list."

    return tfm_list



def warp (x, tfms, tfm_names, backend=None, debug=False):
    '''
    Perform a multi-step transform on the tensor x
    x: tensor
    tfms:  'btd -> b,t,2,d//2 -> b,2,t,d//2 -> b,2,t,^n,d//2'
    tfm_names: 'vp' [first (v)iew, then (p)ermute transform]
    backend:  either a string('numpy', 'tf', 'torch') or the corresponding backend.<class>
    debug: prints per-step debugging information
    '''

    be = get_backend(backend, x)
    tfm_list = tfm_seq_decompose(tfms, tfm_names)
    #print (f'tfm list {tfm_list}')

    ret = x
    for sym, l, r in tfm_list:
        if debug:
            print(f'*** processing transform.. {sym}\n {l} -> {r}')
        if sym == 'v' or sym == 'r': #view transform
            new_shape = _view_transform(l, r, be.shape(ret))
            ret = be.view(x, new_shape)
        elif sym == 'p' or sym == 't':
            perm_indices = _permute_transform(l, r)
            ret = be.transpose(ret, perm_indices)
        elif sym == 'e':
            expand_shape = expand_transform(l, r)
            ret = be.expand(ret, expand_shape)
        elif sym == 'c': 
            ret = be.contiguous(ret)
        elif sym == 'j':
            dims = _join_transform(ret, l, r)
            ret = join(ret, dims, backend=be)
        else:
            assert False, f'Invalid transform symbol {sym}'
        if debug:
            print (f'   after transform shape is: {be.shape(ret)}')
    return ret



from .utils import int_shape

class ABackend:
    name = None

    def shape(self, x): raise NotImplementedError
    def contiguous(self, x): raise NotImplementedError
    def view(self, x, shape): raise NotImplementedError
    def transpose(self, x, dims): raise NotImplementedError
    def expand(self, x, mul_shape): raise NotImplementedError
    def stack(self, x, mul_shape): raise NotImplementedError
    def concat(self, x, mul_shape): raise NotImplementedError


class Numpy(ABackend):
    name = 'numpy'

    def __init__(self):
        import numpy
        self.np = numpy
    def shape(self, x): return x.shape
    def contiguous(self, x): self.np.ascontiguousarray(x)
    def view(self, x, shape): 
        #print (type(x), x.shape, shape)
        return x.reshape(shape)
    def transpose(self, x, dims): return x.transpose(dims)
    def expand(self, x, mul_shape): return x.tile(mul_shape)
    def stack(self, xlist, axis): return self.np.stack(xlist, axis=axis)
    def concat(self, xlist, axis): return self.np.concatenate(xlist, axis=axis)

class PyTorch(ABackend):
    name = 'pytorch'
    def __init__(self):
        import torch
        self.torch = torch
    def shape(self, x): return x.size()
    def contiguous(self, x): return x.contiguous()
    def view(self, x, shape): return x.view(shape)
    def transpose(self, x, dims): return x.permute(dims)
    def expand(self, x, mul_shape): return x.expand(mul_shape)
    def stack(self, xlist, axis): return self.torch.stack(xlist, dim=axis)
    def concat(self, xlist, axis): return self.torch.cat(xlist, dim=axis)


def get_tf_shape(tf, tensor):
  """Returns a list of the shape of tensor, preferring static dimensions.
  (inspired by get_shape_list in BERT code)

  Args:
    tensor: A tf.Tensor object to find the shape of.
  Returns:
    A list of dimensions of the shape of tensor. All static dimensions will
    be returned as python integers, and dynamic dimensions will be returned
    as tf.Tensor scalars.
  """
  shape = tuple(tensor.shape.as_list())

  non_static_indexes = []
  for (index, dim) in enumerate(shape):
    if dim is None:
      non_static_indexes.append(index)

  if not non_static_indexes:
    return shape

  dyn_shape = tf.shape(tensor)
  for index in non_static_indexes:
    shape[index] = dyn_shape[index]
  return shape

class TF(ABackend):
    name = 'tensorflow'
    def __init__(self):
        import tensorflow
        self.tf = tensorflow

    def shape(self, x):
        if self.tf.executing_eagerly():
            return int_shape(x.shape)
        else:
            #return tuple(self.tf.unstack(self.tf.shape(x)))
            return get_tf_shape(self.tf, x)

    def contiguous(self, x): return x

    def view(self, x, shape): return self.tf.reshape(x, shape)
    def transpose(self, x, dims): return self.tf.transpose(x, dims)
    def expand(self, x, mul_shape): return self.tf.tile(x, mul_shape)
    def stack(self, xlist, axis): return self.tf.stack(xlist, axis=axis)
    def concat(self, xlist, axis): return self.tf.concat(xlist, axis=axis)


becache  = {}
def from_cache(C):
    s = str(C)
    if s not in becache:
        becache[s] = C()
    return becache[s]

def get_backend_for_tensor(x):
    '''
    get backend for tensor x
    '''
    t = str(type(x))

    if 'numpy.' in t: ret = from_cache(Numpy)
    elif 'torch.' in t: ret = from_cache(PyTorch)
    elif 'tensorflow.' in t: ret = from_cache(TF)
    else: 
        raise NotImplementedError(f'Unsupported tensor type {t}. Contributions welcome.')

    return ret

def get_backend_by_name (b):
    if isinstance(b, (Numpy, TF, PyTorch)): return b
    assert isinstance(b, str)
    bemap = {
        'numpy': Numpy,
        'np': Numpy,
        'torch': PyTorch,
        'pytorch': PyTorch,
        'tensorflow': TF,
        'tf': TF
    }
    if b in bemap: return from_cache(bemap[b])
    else:
        raise NotImplementedError(f'Unsupported backend {b}. Contributions welcome.')







from .tsn import tsn_to_str_list, tsn_to_tuple


def int_shape(*s):
    if len(s) == 1: 
        assert isinstance(s, (tuple,list))
        s = s[0]
    else: s = tuple(s)
    return tuple([int(d) for d in s])

def select(x, dv_dict, squeeze=False):
    '''
    Index using dimension shorthands
    
    x: (t, 'bcd') -- tensor, shape tuple (can be indexed in numpy notation : x[:,0,:])
    dv_dict: {'b': 0, 'c': 5} 
    squeeze: [True, False] or a tsn list ('b,c') of dims to be squeezed
    '''
    assert not squeeze, 'not implemented'

    assert isinstance(tuple), 'The first argument should be a tuple of (vector, shape)'
    xv, xs = x
    shape, is_seq = tsn_to_str_list(xs)
    if not is_seq:
        raise NotImplementedError(f"get from shape {xs} not implemented")

    colon = slice(None)
    slice_tuple = [colon] * len(shape)
    for pos, sh in shape:
        if sh in dv_dict:
            slice_tuple[pos] = dv_dict[sh]

    y = x[slice_tuple]
    return y

def size_assert(x_size, sa, dims=None):
    '''
    x_size: integer tuple
    sa: TSA
    dims: None or Sequence[int], e.g., [0,1]
    Check if size of tensor x matches TSA `sa` along `dims` axes
    '''
    x_size, sa = tuple(x_size), tuple(sa)
    if dims is not None:
        assert isinstance(dims, (list, tuple))
        x_size = [x_size[d] for d in dims]
        sa = [sa[d] for d in dims]

    if x_size != sa:
        print(f'Size mismatch: size = {x_size}, expected: {sa}')
        assert False


def reduce_dims (tfm):
    '''
    tfm: str, 'btd->b'
    '''
    src, to = tfm.split('->')
    src = tsn_to_tuple(src.strip())
    to = tsn_to_tuple(to.strip())

    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    drops = []
    #check src includes all dims in to
    for i, d in enumerate(src):
        if d not in to:
            drops.append(i)

    return tuple(drops)




name = "tsalib"

#import sys
#if sys.version_info < (3, 6, 1):
#    raise RuntimeError("TSAlib requires Python 3.6.1 or later")

from .ts import dim_var, dim_vars, get_dim_vars
from .tsn import tsn_to_shape
from .utils import select, reduce_dims, size_assert, int_shape
from .transforms import view_transform, permute_transform, expand_transform, join_transform
from .transforms import alignto
from .tensor_ops import warp, join



from .ts import dim_var, DimExpr, dummy_dvar, TupleSeq
from sympy import sympify, Integer
from .tsn import _sexprs_to_ts, tsn_to_str_list, tsn_to_tuple, check_int_tuple, resolve_to_int_tuple



def _view_transform (src, to, in_shape, checkin=False):
    '''
    View Transform
    src, to: Union[str, Tuple]
    :src is the current view of the tensor
    :to is the target view, may contain expressions over named dim variables
    :in_shape is the shape (list/tuple) of the source tensor 
    :returns the new size of the tensor after view transformation (backend independent)
    '''
    if checkin: check_int_tuple(in_shape)

    src = tsn_to_tuple(src)
    if (len(src) != len(in_shape)):
        print (f'{src}, {in_shape}')
        raise ValueError("Source DimExpr does not match input tensor's shape")
    to = tsn_to_tuple(to)
    #print (src, to)
    assert isinstance(in_shape, (list, tuple))
    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    #sub_map = [(d.e, Symbol(f'{i}')) for i, d in enumerate(src)]
    sub_map = [(d.exp, in_shape[i]) for i, d in enumerate(src)]
    out_shape = tuple([t.exp.subs(sub_map) if isinstance(t, DimExpr) else int(t) for t in to])

    out_shape = resolve_to_int_tuple(out_shape)
    return out_shape

def view_transform (tfm, in_shape):
    '''
    View transform
    :tfm:str is the shorthand representation of the transform ('btd -> b,t*2,d//2')
    '''
    l, r = tfm.split('->')
    return _view_transform(l.strip(), r.strip(), in_shape)


def _permute_transform(src, to):
    '''
    Permute Transform
    src, to: Union[str, Tuple]

    :src is the current dimension arrangement, list of named dim variables
    :to is the target dimension arragement, list of named dim variables
    :returns the index tuple for the permutation (backend independent)
    '''
    src = tsn_to_tuple(src)
    to = tsn_to_tuple(to)
    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    assert len(src) == len(to), "Source and Target shapes for permutation are not same"

    sub_map = [(d.exp, i) for i, d in enumerate(src)]
    perm_indices = tuple([t.exp.subs(sub_map) for t in to])
    perm_indices = resolve_to_int_tuple(perm_indices)

    return perm_indices


def permute_transform (tfm):
    '''
    Permute transform
    :tfm: str
    :tfm is the shorthand representation of the transform (',t,d -> ,d,t')
    '''
    l, r = tfm.split('->')
    return _permute_transform(l.strip(), r.strip())





def _join_transform (tlist, src, to):
    '''
    src: Union[str, TupleSeq]
    to: Union[str, tuple]
    src: (b,c,d)* , to: '^, b, c, d'    OR
    src: (b,c,d)* , to: 'b, 3*c, d'

    returns the dims shorthand for joining (backend independent)

    '''
    lhs = tsn_to_tuple(src) #TupleSeq(B, C, D)
    rhs = tsn_to_tuple(to) 

    assert isinstance(lhs, TupleSeq)
    assert isinstance(rhs, tuple)

    lhs = lhs.item() # (b, c, d)
    int1 = Integer(1) # substitute all dim shorthands by '1'

    sub_map = [(d.exp, int1) for d in lhs]
    dims = tuple([t.exp.subs(sub_map) for t in rhs]) # (^, 1, 1, 1)


    if len(rhs) == len(lhs): #concat, join by '*'
        #join_dims = (1,3*1,1)
        dims = ','.join(map(lambda x: '' if x == int1 else '*', dims))
        #dims = ',*,'
    elif len(rhs) == (len(lhs) + 1): #stack, join by '^'
        dims = ','.join(map(lambda x: '' if x == int1 else '^', dims))
        #dims = '^,,,'
    else:
        raise ValueError(f'Unable to join from {src} to {to}')

    return dims


def join_transform (tlist, tfm):
    '''
    Join transform (backend independent)
    :tlist: List[tensor]
    :tfm:str represents the transform "(b,c,d)* -> ^,b,c,d"
    returns the dims in TSN for joining 

    '''
    l, r = tfm.split('->')
    return _join_transform(tlist, l.strip(), r.strip())


def align_transform (src, to):
    '''
    src: tsn 'd,d'
    to: tsn '6, d, t, d'
    return: '^,,^,' [expansion shorthand for src]
    '''

    lhs = tsn_to_tuple(src) #TupleSeq(B, C, D)
    rhs = tsn_to_tuple(to) 

    assert isinstance(lhs, tuple)
    assert isinstance(rhs, tuple)

    lhs_pos, rhs_pos = 0, 0
    res = []
    for rhs_pos, d in enumerate(rhs):
        if lhs[lhs_pos] == d: 
            #print ('match', d, lhs[lhs_pos])
            res.append('')
            lhs_pos += 1
        else:
            res.append('^')

    #print (lhs_pos, res)

    if lhs_pos != len(lhs):
        print (f'Unable to align {src} to {to}: {src} not a subsequence of {to}')
        raise ValueError

    return ','.join(res)



def _expansions_as_dict(expansions):
    if isinstance(expansions, list): #[(T, T*5), (D, D*4)]
        res = expansions
    else:
        assert isinstance(expansions, str) #'k->k*5,t->t*10'
        expansions = expansions.strip().split(',')
        res = []
        for ex in expansions:
            t = _sexprs_to_ts(ex.strip().split('->'))
            #print(t)
            res.append(t)
            #print (res)

    exp_map = {l: r for (l, r) in res}
    return exp_map


def expand_transform (src, expansions, in_shape):
    '''
    :src        (B, T, D) = (10, 20, 300)
    :expansions [(T, T*5), (D, D*4)]
    :returns the expansion shape tuple
    '''
    src = tsn_to_tuple(src)
    exp_map = _expansions_as_dict(expansions)
    #exp_map = {_sexpr_to_ts(s)[0]: _sexpr_to_ts(t)[0] for (s, t) in (expansions)}
    sub_map = [(d.exp, in_shape[i]) for i, d in enumerate(src)] # (B, 10), (T, 20), (D, 300)

    #print (expansions, exp_map)

    res = []
    for k in src:
        if k not in exp_map: res.append(-1) #keep dim shape same
        else:
            v = exp_map[k]
            assert isinstance(v, DimExpr)
            res.append(v.exp.subs(sub_map)) #(T*5) -> 100, (D*4) -> 1200

    res = tuple(res)
    res = resolve_to_int_tuple(res)
    return res

def expand_transform2(x, tfm):
    '''
    x: (backend) tensor, e.g., shape 'd,d'
    tfm: expand tsn: '^,,^,'
    returns tensor of shape '1,d,1,d'
    '''

    #print (f'expand: {tfm}')
    colon = slice(None)
    expand_tup = tuple(None if c == '^' else colon for c in tfm.split(','))
    return x[expand_tup]


def alignto(x, ys, expand=False):
    '''
    Align tensor x's shape to y's shape.
    Assume x's shape is a subsequence of y's shape
    Assume tensors x and y support numpy's "None, :"" indexing notation
    x: (tensor_var, x_shape)
    ys: y_shape (tsn of target tensor)
    '''

    assert isinstance(x, tuple), 'First argument is of form (tensor_var, tsn)'
    xt, xs = x
    expand_tfm = align_transform(xs, ys)
    return expand_transform2(xt, expand_tfm)
















class TSLite:
    def __init__(self, name, w=1, b=0):
        self.name = name
        self.w = w
        self.b = b

    def __add__(self, n):
        if isinstance(n, int):
            return TSLite(self.name, self.w, self.b + n)
        elif isinstance(n, TSLite):
            return [self, n]
        else:
            assert False

    def __mul__(self, n):
        if isinstance(n, int):
            return TSLite(self.name, self.w * n, self.b)
        else:
            assert False

    def __div__(self, n):
        if isinstance(n, int):
            return TSLite(self.name, self.w / n, self.b)
        else:
            assert False

    def __eq__(self, d):
        return self.name == d.name and self.w == d.w and self.b == d.b

    def __repr__(self):
        s = f'({self.name}*{self.w}+{self.b})'
        return s

